# socketify.py
Fast WebSocket and Http/Https server using CFFI with C API from [uNetworking/uWebSockets](https://github.com/uNetworking/uWebSockets)

> This project will adapt the full capi from uNetworking/uWebSockets but for now it's just this.

### Overly simple hello world app
```python
from socketify import App

app = App()
app.get("/", lambda res, req: res.end("Hello World socketify from Python!"))
app.listen(3000, lambda config: print("Listening on port http://localhost:%d now\n" % config.port))
app.run()
```

### pip install

```bash
pip install git+https://github.com/cirospaciari/socketify.py.git --global-option=build_ext
#or specify PyPy3
pypy3 -m pip install git+https://github.com/cirospaciari/socketify.py.git --global-option=build_ext
```

### Run
```bash
pypy3 ./hello_world.py
```

### SSL version sample
``` python
from socketify import App, AppOptions

app = App(AppOptions(key_file_name="./misc/key.pem", cert_file_name="./misc/cert.pem", passphrase="1234"))
app.get("/", lambda res, req: res.end("Hello World socketify from Python!"))
app.listen(3000, lambda config: print("Listening on port http://localhost:%d now\n" % config.port))
app.run()
```

### Build local from source
```bash
#clone and update submodules
git clone https://github.com/cirospaciari/socketify.py.git
cd ./socketify.py
git submodule update --init --recursive --remote
#install build with pip
pypy3 -m pip install --upgrade build
#build and install
pypy3 -m build
pypy3 -m pip install . --no-cache-dir --global-option=build_ext
#if you want to remove
pypy3 -m pip uninstall socketify
```